//
//  TRBuffer.h
//
//  Created by Grégoire Milliez on 07/01/2014.
//
//  This class is using circular_buffer from boost
//  to implement a timed ring buffer (TRBuffer).
 
#ifndef TRBUFFER_H
#define TRBUFFER_H
 
#include <boost/circular_buffer.hpp>

template <typename T> class TRBuffer {
    boost::circular_buffer<int> m_TimeBuffer;
    boost::circular_buffer<T> m_DataBuffer;

public:

    TRBuffer(){
		m_TimeBuffer.set_capacity(100);
		m_DataBuffer.set_capacity(100);
	}

	TRBuffer(const size_t capacity){
		m_TimeBuffer.set_capacity(capacity);
		m_DataBuffer.set_capacity(capacity);
	}
 
	~TRBuffer() {}

    // Get the number of elements currently stored in the TRBuffer.
	boost::circular_buffer<int>::size_type size(){ m_TimeBuffer.size(); }
	// Get the maximum number of elements stored in the TRBuffer.
	boost::circular_buffer<int>::size_type capacity(){ m_TimeBuffer.capacity(); }
	boost::circular_buffer<int>::size_type max_size(){ m_TimeBuffer.max_size(); }
	bool empty(){ m_TimeBuffer.empty(); }
	bool full(){ m_TimeBuffer.full(); }

	// Insert a new element at the end of the TRBuffer.
	// If TRBuffer is full, the first element will be removed. 
	// If the capacity is 0, nothing will be inserted.
	void push_back(int time, const T &data){
	  m_TimeBuffer.push_back(time);
	  m_DataBuffer.push_back(data);
	}

    // Remove the first element from the TRBuffer. 
    void pop_front(){
	  m_TimeBuffer.pop_front();
	  m_DataBuffer.pop_front();
	}
    
    // Remove all stored elements from the TRBuffer. 
    void clear(){
	  m_TimeBuffer.clear();
	  m_DataBuffer.clear();
	}

    bool getData(int time, T &data){
      int result;
      result = getIndex(time);
      if (result < 0)
        return false;
      else{
        data = getDataFromIndex( result );
        return true;
      }
    }

    // Here after means strictly greater time.
    bool getDataAfter(int time, T &data){
      int result;
      result = getIndexAfter(time);
      if (result < 0)
        return false;
      else{
        data = getDataFromIndex( result );
        return true;
      }
    }
    
    // Here before means last data before this time.
    bool getDataBefore(int time, T &data){
      int result;
      result = getIndexBefore(time);
      if (result < 0)
        return false;
      else{
        data = getDataFromIndex( result );
        return true;
      }
    }

    int getIndex(int time){
      if( !m_TimeBuffer.empty() ){
        for(int i=0; i < m_TimeBuffer.size() && i<=time; i++){
          if(m_TimeBuffer[i] == time)
            return i;
          else if(m_TimeBuffer[i] > time){
            printf("No data was found at time %d\n", time);
            return -1;
          }
        }
      }
    }

    // Here after means strictly greater time.
    int getIndexAfter(int time){
      if( !m_TimeBuffer.empty() ){
        for(int i=0; i<m_TimeBuffer.size(); i++){
          if(m_TimeBuffer[i] > time)
            return i;
        }
        printf("No data was found after time %d\n", time);
        return -1;
      }
    }
    
    // Here before means last data before this time.
    int getIndexBefore(int time){
      if( !m_TimeBuffer.empty() ){
        for(int i=0; i<m_TimeBuffer.size(); i++){
          if(m_TimeBuffer[i] > time)
            return i-1;
        }
        printf("Warning, input time is greater than all data in Ring buffer %d\n", time);
        return m_TimeBuffer.size()-1;
      }
    }

  T getDataFromIndex(int index){
    return m_DataBuffer[index];
  }
};

//#include "TRBuffer.hpp"
#endif
